﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;

namespace HeuristicLab.Problems.PTSP {
  [Item("PTSPEstimatedInsertionEvaluator", "Evaluates an insertion move (1-shift)")]
  [StorableClass]
  public class PTSPEstimatedInsertionEvaluator : PTSPMoveEvaluator, IPermutationTranslocationMoveOperator {

    public override Type EvaluatorType {
      get { return typeof(PTSPEstimatedInsertionEvaluator); }
    }
    public ILookupParameter<TranslocationMove> TranslocationMoveParameter {
      get { return (ILookupParameter<TranslocationMove>)Parameters["TranslocationMove"]; }
    }

    public IValueParameter<ItemList<ItemList<IntValue>>> RealizationsParameter {
      get { return (IValueParameter<ItemList<ItemList<IntValue>>>)Parameters["Realizations"]; }
    }

    [StorableConstructor]
    protected PTSPEstimatedInsertionEvaluator(bool deserializing) : base(deserializing) { }
    protected PTSPEstimatedInsertionEvaluator(PTSPEstimatedInsertionEvaluator original, Cloner cloner) : base(original, cloner) { }
    public PTSPEstimatedInsertionEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<TranslocationMove>("TranslocationMove", "The move to evaluate."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PTSPEstimatedInsertionEvaluator(this, cloner);
    }

    public static double EvaluateByCoordinates(Permutation permutation, TranslocationMove move, DoubleMatrix coordinates, PTSPEstimatedInsertionEvaluator evaluator) {
      int edge1source = permutation.GetCircular(move.Index1 - 1);
      int edge1target = permutation[move.Index1];
      int edge2source = permutation[move.Index2];
      int edge2target = permutation.GetCircular(move.Index2 + 1);
      if (move.Index2 - move.Index1 >= permutation.Length - 2) return 0;
      double moveQuality = 0;
      // remove two edges
      moveQuality -= evaluator.CalculateDistance(coordinates[edge1source, 0], coordinates[edge1source, 1],
            coordinates[edge1target, 0], coordinates[edge1target, 1]);
      moveQuality -= evaluator.CalculateDistance(coordinates[edge2source, 0], coordinates[edge2source, 1],
        coordinates[edge2target, 0], coordinates[edge2target, 1]);
      // add two edges
      moveQuality += evaluator.CalculateDistance(coordinates[edge1source, 0], coordinates[edge1source, 1],
        coordinates[edge2source, 0], coordinates[edge2source, 1]);
      moveQuality += evaluator.CalculateDistance(coordinates[edge1target, 0], coordinates[edge1target, 1],
        coordinates[edge2target, 0], coordinates[edge2target, 1]);
      return moveQuality;
    }

    public static double EvaluateByDistanceMatrix(Permutation permutation, TranslocationMove move, DistanceMatrix distanceMatrix, ItemList<ItemList<IntValue>> realizations) {
      Permutation afterMove = new Permutation(PermutationTypes.RelativeUndirected,permutation);
      TranslocationManipulator.Apply(afterMove, move.Index1, move.Index1, move.Index3);
      double moveQuality = 0;
      int[] edges = new int[12];
      int[] indices = new int[12];
      edges[0] = permutation.GetCircular(move.Index1 - 1);
      indices[0] = decreaseCircularIndex(permutation.Length, move.Index1);
      edges[1] = permutation[move.Index1];
      indices[1] = move.Index1;
      edges[2] = permutation[move.Index1];
      indices[2] = move.Index1;
      edges[3] = permutation.GetCircular(move.Index1 + 1);
      indices[3] = increaseCircularIndex(permutation.Length, move.Index1);

      edges[6] = afterMove.GetCircular(move.Index3 - 1);
      indices[6] = decreaseCircularIndex(afterMove.Length, move.Index3);
      edges[7] = afterMove[move.Index3];
      indices[7] = move.Index3;
      edges[8] = afterMove[move.Index3];
      indices[8] = move.Index3;
      edges[9] = afterMove.GetCircular(move.Index3 + 1);
      indices[9] = increaseCircularIndex(afterMove.Length, move.Index3);

      if (move.Index3 > move.Index1) {
        edges[4] = permutation[move.Index3];
        indices[4] = move.Index3;
        edges[5] = permutation.GetCircular(move.Index3 + 1);
        indices[5] = indices[9];
        edges[10] = afterMove.GetCircular(move.Index1 - 1);
        indices[10] = indices[0];
        edges[11] = afterMove[move.Index1];
        indices[11] = move.Index1;
      } else {
        edges[4] = permutation.GetCircular(move.Index3 - 1);
        indices[4] = indices[6];
        edges[5] = permutation[move.Index3];
        indices[5] = move.Index3;
        edges[10] = afterMove[move.Index1];
        indices[10] = move.Index1;
        edges[11] = afterMove.GetCircular(move.Index1 + 1);
        indices[11] = indices[3];
      }
      int[] aPosteriori = new int[12];
      Permutation tempPermutation;
      foreach (ItemList<IntValue> realization in realizations) {
        for (int i = 0; i < edges.Length; i++) {
          if (i < 6) {
            tempPermutation = permutation;
          } else {
            tempPermutation = afterMove;
          }
          if (realization[edges[i]].Value == 1) {
            aPosteriori[i] = edges[i];
          } else {
            int j=1;
            if (i%2==0) {
              // find nearest predecessor in realization if source edge
              while (realization[tempPermutation.GetCircular(indices[i] - j)].Value == 0) {
                j++;
              }
              aPosteriori[i] = tempPermutation.GetCircular(indices[i] - j);
            } else {
              // find nearest successor in realization if target edge
              while (realization[tempPermutation.GetCircular(indices[i] + j)].Value == 0) {
                j++;
              }
              aPosteriori[i] = tempPermutation.GetCircular(indices[i] + j);
            }
          }
        }
        if (!(aPosteriori[0] == aPosteriori[2] && aPosteriori[1] == aPosteriori[3]) &&
          !(aPosteriori[0] == aPosteriori[4] && aPosteriori[1] == aPosteriori[5]) &&
          !(aPosteriori[2] == aPosteriori[4] && aPosteriori[3] == aPosteriori[5])) {
          // compute cost difference between the two a posteriori solutions
          moveQuality = moveQuality + distanceMatrix[aPosteriori[6], aPosteriori[7]] + distanceMatrix[aPosteriori[8], aPosteriori[9]] + distanceMatrix[aPosteriori[10], aPosteriori[11]];
          moveQuality = moveQuality - distanceMatrix[aPosteriori[0], aPosteriori[1]] - distanceMatrix[aPosteriori[2], aPosteriori[3]] - distanceMatrix[aPosteriori[4], aPosteriori[5]];
        }
        Array.Clear(aPosteriori, 0, aPosteriori.Length);
      }
      // return average of cost differences
      return moveQuality / realizations.Capacity;
    }

    private static int decreaseCircularIndex(int length, int index) {
      int result = index - 1;
      if (result == -1) {
        result = length - 1;
      }
      return result;
    }

    private static int increaseCircularIndex(int length, int index) {
      int result = index + 1;
      if (result == length + 1) {
        result = 0;
      }
      return result;
    }

    protected override double EvaluateByCoordinates(Permutation permutation, DoubleMatrix coordinates) {
      return EvaluateByCoordinates(permutation, TranslocationMoveParameter.ActualValue, coordinates, this);
    }

    protected override double EvaluateByDistanceMatrix(Permutation permutation, DistanceMatrix distanceMatrix) {
      return EvaluateByDistanceMatrix(permutation, TranslocationMoveParameter.ActualValue, distanceMatrix, RealizationsParameter.Value);
    }

    protected override double CalculateDistance(double x1, double y1, double x2, double y2) {
      return Math.Round(Math.Sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)));
    }
  }
}
